from typing import Annotated
from contextlib import asynccontextmanager
from importlib.metadata import entry_points
import logging

from pydantic import BaseModel, ValidationError
from fastapi import FastAPI, Body, HTTPException, status
from fastapi.middleware.cors import CORSMiddleware

from jwt import decode as decode_jwt

from knoc import __version__
from knoc import workflows
from knoc.settings import settings
from knoc.models import ServerInfo, APIPlugin, WorkflowPlugin
from knoc.workflows import workflows
from knoc.hatchet_client import hatchet
from hatchet_sdk.clients.rest.models.worker import Worker

# from knoc.workflows.utils import ImportReport
# from knoc.db import get_session

logger = logging.getLogger("knoc.server")


@asynccontextmanager
async def lifespan(app: FastAPI):
    # Load the FastAPI app
    assert app
    logger.debug("App is Starting Up!")
    yield
    logger.info(
        "App is Shutting Down Gracefully!",
        extra={"markup": True},
    )


app = FastAPI(lifespan=lifespan)
app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.server.cors,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

server_info = ServerInfo(
    version=__version__,
    hatchet_url=decode_jwt(
        settings.hatchet_client.token, options={"verify_signature": False}
    )["server_url"],
    api_doc="/docs",
)


def get_plugins():
    for entry_point in entry_points().select(group="knoc.api"):
        try:
            api = entry_point.load()
        except Exception as err:
            logger.warning(
                f"Cannot load settings module {entry_point.name} - see below"
            )
            logger.exception(err)
            continue
        if not isinstance(api, FastAPI):
            logger.warning(f"Settings module {entry_point.name} is not a FastAPI")
            continue
        app.mount("/" + entry_point.name, api)
        server_info.plugins.api[entry_point.value] = APIPlugin(
            id=entry_point.name,
            module=entry_point.module,
            description=api.description,
            api=api,
        )
        logger.info(f"Mounted {entry_point.name}")


server_info.plugins.workflow = workflows

get_plugins()


@app.get("/")
async def root() -> ServerInfo:
    return server_info


@app.get("/wf")
async def list_workflows() -> dict[str, WorkflowPlugin]:
    """List the workflows"""
    # return {k: wf.wf.config for k, wf in workflows.items()}  # type: ignore
    return workflows


@app.post("/wf/{workflow_id}")
async def run_workflow(workflow_id: str, input: Annotated[dict, Body()]):
    """Execute a workflow"""
    wf = workflows[workflow_id].wf
    try:
        input_ = wf.config.input_validator(**input)
    except ValidationError as err:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=err.title)
    else:
        result = await wf.aio_run(input=input_)
    return result


@app.get("/worker")
async def list_workers() -> list[Worker] | None:
    """List the workers"""
    workers = await hatchet.workers.aio_list()
    return workers.rows
